#include <windows.h>
#include <shlwapi.h>

#include <stdlib.h>

#include "board/sg-reader.h"
#include "board/io4.h"

#include "hook/process.h"

#include "hooklib/dvd.h"
#include "hooklib/serial.h"
#include "hooklib/spike.h"

#include "idachook/config.h"
#include "idachook/idac-dll.h"
#include "idachook/io4.h"
#include "idachook/zinput.h"

#include "platform/platform.h"

#include "util/dprintf.h"

static HMODULE idac_hook_mod;
static process_entry_t idac_startup;
static struct idac_hook_config idac_hook_cfg;

static DWORD CALLBACK idac_pre_startup(void)
{
    HRESULT hr;

    dprintf("--- Begin idac_pre_startup ---\n");

    /* Config load */

    idac_hook_config_load(&idac_hook_cfg, L".\\segatools.ini");

    /* Hook Win32 APIs */

    serial_hook_init();
    zinput_hook_init(&idac_hook_cfg.zinput);
    dvd_hook_init(&idac_hook_cfg.dvd, idac_hook_mod);

    /* Initialize emulation hooks */

    hr = platform_hook_init(
            &idac_hook_cfg.platform,
            "SDGT",
            "ACA4",
            idac_hook_mod);

    if (FAILED(hr)) {
        goto fail;
    }

    hr = sg_reader_hook_init(&idac_hook_cfg.aime, 3, 3, idac_hook_mod);

    if (FAILED(hr)) {
        goto fail;
    }

    hr = idac_dll_init(&idac_hook_cfg.dll, idac_hook_mod);

    if (FAILED(hr)) {
        goto fail;
    }

    hr = idac_io4_hook_init(&idac_hook_cfg.io4);

    if (FAILED(hr)) {
        goto fail;
    }

    /* Initialize debug helpers */

    spike_hook_init(L".\\segatools.ini");

    dprintf("---  End  idac_pre_startup ---\n");

    /* Jump to EXE start address */

    return idac_startup();

fail:
    ExitProcess(EXIT_FAILURE);
}

BOOL WINAPI DllMain(HMODULE mod, DWORD cause, void *ctx)
{
    HRESULT hr;

    if (cause != DLL_PROCESS_ATTACH) {
        return TRUE;
    }

    idac_hook_mod = mod;

    hr = process_hijack_startup(idac_pre_startup, &idac_startup);

    if (!SUCCEEDED(hr)) {
        dprintf("Failed to hijack process startup: %x\n", (int) hr);
    }

    return SUCCEEDED(hr);
}
